# Copyright (c) 2023 DomainTools LLC
# Copyright (c) 2012-2016, 2018-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import collections.abc
import json
import os
import ipaddress

CONF_FILE = os.environ.get('DNSTLI_CONFIG_FILE', '/etc/dnstli/dnstli2.conf')
API_DIR = '/etc/dnstli/auth-api'

# for minimum and maximum version number checks
MY_API_VERSION = 2

DEFAULT_QUERY_EXIT_CHECK_INTERVAL = 0.25
DEFAULT_MAX_QUERY_TIME = 999999     # very long number of seconds
DEFAULT_OFFSET_MAX = 999999999999   # very large number of rows

required_config_keys = (
    'filesets',
    'res_limit_default',
    'res_limit_max',
)

optional_config_keys = (
    'admin_addresses', # list of addresses that may connect as any account
    'export_installation', # true if an export installation
    'query_limit_default',
    'query_limit_period',
    'query_limit_redis_socket',
    'query_limit_redis_url',
    'query_exit_check_interval',
    'max_query_time',
    'offset_max',
    'cors_origins',
    'heartbeat_interval',
    'wildcard_label'
)

positive_integer_config_keys = (
    'res_limit_default',
    'res_limit_max',
    'offset_max',
    'query_limit_default',
    'query_limit_period',
)

nullable_positive_float_config_keys = (
    'query_exit_check_interval',
    'max_query_time',
    'heartbeat_interval',
)

def get_config(config_fname=None):
    '''
    Returns a configuration dictionary, with the above required_config_keys and possibly optional_config_keys.
    Raises exceptions with any problems it finds in the config.
    '''
    if config_fname is None:
        config_fname = CONF_FILE

    with open(config_fname) as f:
        print("loading config from file {}\n".format(config_fname))
        try:
            config = json.load(f)
        except Exception as e:
            raise Exception("loading config from file {} failed with {}\n".format(config_fname,str(e)))

    for k in required_config_keys:
        if k not in config:
            raise Exception('config key "%s" is required but not present in config file %s' % (k, config_fname))

    for k in config:
        if not (k in required_config_keys or k in optional_config_keys):
            raise Exception('config key "%s" is unknown (and thus unexpected) in config file %s' % (k, config_fname))

    for k in positive_integer_config_keys:
        if k in config:
            try:
                config[k] = int(config[k])
            except ValueError:
                raise Exception('config key "%s" is not an integer in config file %s' % (k, config_fname))
            if config[k] < 0:
                raise Exception('config key "%s" is a negative integer in config file %s' % (k, config_fname))

    for k in nullable_positive_float_config_keys:
        if k in config:
            try:
                config[k] = float(config[k])
            except ValueError:
                raise Exception('config key "%s" is not a float in config file %s' % (k, config_fname))
            if config[k] < 0:
                raise Exception('config key "%s" is a negative float in config file %s' % (k, config_fname))

    # if these are missing, then use defaults
    if config.get('query_exit_check_interval') is None:
        config['query_exit_check_interval'] = DEFAULT_QUERY_EXIT_CHECK_INTERVAL
    if config.get('max_query_time') is None:
        config['max_query_time'] = DEFAULT_MAX_QUERY_TIME
    if config.get('offset_max') is None:
        config['offset_max'] = DEFAULT_OFFSET_MAX
    if 'wildcard_label' not in config:
        config['wildcard_label'] = None

    if 'query_limit_redis_url' in config and 'query_limit_redis_socket' in config:
        raise Exception('query_limit_redis_url and query_limit_redis_socket are mutually exclusive in config file %s' % config_fname)

    if 'query_limit_redis_url' in config and not ('query_limit_default' in config and 'query_limit_period' in config):
        raise Exception('setting query_limit_redis_url requires also setting query_limit_default, query_limit_period in config file %s' % config_fname)

    if 'query_limit_redis_socket' in config and not ('query_limit_default' in config and 'query_limit_period' in config):
        raise Exception('setting query_limit_redis_socket requires also setting query_limit_default, query_limit_period in config file %s' % config_fname)

    if not isinstance(config.get('cors_origins', []), collections.abc.Iterable) or isinstance(config.get('cors_origins', []), str):
        raise Exception('cors_origins must be iterable in config file %s' % config_fname)

    admin_addresses = config.get('admin_addresses', [])
    if not isinstance(admin_addresses, collections.abc.Iterable) or isinstance(admin_addresses, str):
        raise Exception('admin_addresses must be iterable in config file %s' % config_fname)

    for acl in admin_addresses:
        try:
            # if acl is not expressed as a network, it will be treated as a
            # /32 (IPv4) or /128 (IPv6).  This will fail if expressed as a
            # network but host bits are set.
            ipaddress.ip_network(acl)
        except ValueError as e:
            raise Exception('Invalid admin acl %s: %s in config file %s' % (acl, e, config_fname))

    if 'export_installation' in config: # optional config parameter but if present, must be a bool
        ei = config.get('export_installation')
        if not isinstance(ei, bool):
            raise Exception("'export_installation' value %s is not a boolean (True or False) in config file %s"
                            % (ei, config_fname))

        if ei and ('query_limit_redis_socket' in config or 'query_limit_redis_url' in config):
            raise Exception("An export_installation may not have redis config parameters defined, in config file %s"
                            % (config_fname))

    return config
