# Copyright (c) 2023 DomainTools LLC
# Copyright (c) 2015, 2019-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import time
from typing import Optional


def parse_time_specifier(spec: Optional[str], now=None, val_err_ok=True) -> Optional[int]:
    if spec is None:
        return None
    if now is None:
        now = int(time.time())

    try:
        specint = int(spec)
    except ValueError as e:
        if val_err_ok:
            return None
        raise e
    except Exception:
        return None

    if spec[0] in '+-':
        return now + specint
    else:
        return specint


class TimeFilter(object):
    fields = (
        'time_first_after',
        'time_first_before',
        'time_last_after',
        'time_last_before',
    )

    def __init__(self, now=None):
        if now is None:
            now = int(time.time())
        self.now = int(now)

        for field in TimeFilter.fields:
            setattr(self, field, None)

    def apply_query_string(self, qs):
        for field in TimeFilter.fields:
            try:
                spec = parse_time_specifier(qs.get(field, [None])[0], self.now, val_err_ok=False)
            except ValueError as e:
                raise e
            setattr(self, field, spec)

    def _apply_policy_qp_set(self, val, field):
        spec = parse_time_specifier(val, self.now)
        setattr(self, field, spec)

    def _apply_policy_qp_min(self, val, field):
        spec = parse_time_specifier(val, self.now)
        old_spec = getattr(self, field)
        if (old_spec is None) or (old_spec is not None and old_spec < spec):
            setattr(self, field, spec)

    def _apply_policy_qp_max(self, val, field):
        spec = parse_time_specifier(val, self.now)
        old_spec = getattr(self, field)
        if (old_spec is None) or (old_spec is not None and old_spec > spec):
            setattr(self, field, spec)

    def apply_policy(self, policy):
        if policy is None or not isinstance(policy, dict):
            return

        # Process "qp_min" policy. These values set floors that query string
        # parameters sent by the client must not exceed. If the client does not
        # send a given query string parameter, and one is specified by the
        # policy, the one specified by the policy is loaded.
        qp_min = policy.get('qp_min', None)
        if qp_min:
            for field in TimeFilter.fields:
                if field in qp_min:
                    self._apply_policy_qp_min(qp_min[field], field)

        # Process "qp_max" policy. These values set ceilings that query string
        # parameters sent by the client must not exceed. If the client does not
        # send a given query string parameter, and one is specified by the
        # policy, the one specified by the policy is loaded, unless a previous
        # "qp_min" policy for this field was set.
        qp_max = policy.get('qp_max', None)
        if qp_max:
            for field in TimeFilter.fields:
                if field in qp_max:
                    self._apply_policy_qp_max(qp_max[field], field)

        # Process "qp_set" policy. These values override any query string
        # parameters set by the client, and they override any other policy.
        qp_set = policy.get('qp_set', None)
        if qp_set:
            for field in TimeFilter.fields:
                if field in qp_set:
                    self._apply_policy_qp_set(qp_set[field], field)
