# Copyright (c) 2012-2020 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import errno
import itertools
import json
import re
import urllib.parse
import ipaddress

from dnstli2.log import *
import dnstli2.conf
import dnstli2.fmt
import dnstli2.rate
import dnstli2.response
import dnstli2.util

# Get our configuration dictionary.  See dnstli2.conf for the required_config_keys and optional_config_keys.
# get_config() may raise an error, which we let kill the process
conf = dnstli2.conf.get_config()

# Only accept API keys that match this pattern.
# Previously passed re.IGNORECASE flag, but that is wrong since the API keys become case-sensitive file names
apikey_pattern = re.compile('^[-a-z0-9]+$')

def get_rate_limiter():
    '''
      If redis is not configured in the dnstli.conf file then all quotas are unlimited.
      If redis is configured:
        If there is no quota defined for a particular API key
        or if redis is unavailable upon startup
        or if redis fails after startup and it throws a ConnectionError or TimeoutError
        then quotas are treated as unlimited (and queries work).
        Once redis recovers or a quota is defined for the key then quotas will work normally.
        Other types of failures from redis will return a 500 Server Error.
      '''
    if conf.get('export_installation') is True:
        return None             # unlimited but don't print anything about it
    import redis
    if 'query_limit_redis_url' in conf:
        redis_socket = redis.from_url(conf['query_limit_redis_url'])
        return dnstli2.rate.RateLimiter(redis_socket, conf['query_limit_period'])
    elif 'query_limit_redis_socket' in conf:
        redis_socket = redis.Redis(unix_socket_path=conf['query_limit_redis_socket'])
        return dnstli2.rate.RateLimiter(redis_socket, conf['query_limit_period'])
    else:
        print("Not configured to use redis - will use unlimited quotas")
        return None

def log_rate_limit_exception(env):
    msg = dnstli2.util.form_basic_log(env, 'rate-limit-exception')
    #$$$ For groups: will need to determine if also exceeded the group's quota and report that
    log(NOTICE, json.dumps(msg))

def log_rate_quota_expired(env):
    ''' $$$ for now, the same message as for a rate_limit_exception '''
    msg = dnstli2.util.form_basic_log(env, 'rate-limit-exception')
    #$$$ For groups: will need to determine if also exceeded the group's quota and report that
    log(NOTICE, json.dumps(msg))

def autoq_env2RateParamsUser(env):
    ''' Generate rate params tuple for a User quota '''
    if env is None:
        return dnstli2.rate.allNoneAutoQRateParams
    maximum = env.get('dnstli.limit')
    if not maximum:
        maximum = conf['query_limit_default']
    return dnstli2.rate.AutoQRateParams(authtype = env.get('dnstli.authtype'),
                                       identifier = env.get('dnstli.authuser'),
                                       maximum = maximum,
                                       period = env.get('dnstli.period'))

def autoq_env2RateParamsGroup(env):
    ''' Generate rate params tuple for a Group quota '''
    if env is None:
        return dnstli2.rate.allNoneAutoQRateParams
    return dnstli2.rate.AutoQRateParams(authtype = 'group',
                                       identifier = env.get('dnstli.authgroup'),
                                       maximum = env.get('dnstli.group_limit'),
                                       period = env.get('dnstli.group_period'))

def blkq_env2RateParamsUser(env):
    ''' Generate rate params tuple for a User quota '''
    if env is None:
        return dnstli2.rate.allNoneBlkQRateParams
    return dnstli2.rate.BlkQRateParams(authtype = env.get('dnstli.authtype'),
                                      identifier = env['dnstli.authuser'],
                                      blkq_split = env.get('dnstli.blkq_split'),
                                      blkq_baseline_used = env.get('dnstli.blkq_baseline_used'),
                                      blkq_expires = env.get('dnstli.blkq_expires'),
                                      burst_size = env.get('dnstli.burst_size'),
                                      burst_window = env.get('dnstli.burst_window'))

class ratelimitmixin:
    '''
    a base class that requires self.rate to be defined in a subclass
    if either of the base class' methods are called.

    We call the *_limit methods with a list of RateParams namedtuples.

    When creating a namedtuple, we must provide all of the defined elements.
    In our case, some of them will always be None, which is ok.
    '''
    def autoq_get_limit(self, env):
        if env is None or self.rate is None:
            return dnstli2.rate.unlimited
        try:
            return self.rate.autoq_get_limit([ autoq_env2RateParamsUser(env), autoq_env2RateParamsGroup(env) ])
        except dnstli2.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return dnstli2.rate.unlimited
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

    def autoq_incr_limit(self, env):
        '''
        This will raise dnstli2.response.RateLimitExceeded if we exceed the maximum.
        '''
        if env is None or self.rate is None:
            return dnstli2.rate.unlimited
        try:
            return self.rate.autoq_incr_limit([ autoq_env2RateParamsUser(env), autoq_env2RateParamsGroup(env) ])
        except dnstli2.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return dnstli2.rate.unlimited
        except dnstli2.rate.LimitExceeded:
            log_rate_limit_exception(env)
            raise dnstli2.response.RateLimitExceeded
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

    def blkq_get_limit(self, env):
        if env is None or self.rate is None:
            return dnstli2.rate.unlimited
        try:
            return self.rate.blkq_get_limit([ blkq_env2RateParamsUser(env) ])
        except dnstli2.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return dnstli2.rate.unlimited
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

    def blkq_incr_limit(self, env):
        '''
        This will raise dnstli2.response.RateLimitExceeded if we exceed the maximum.
        '''
        if env is None or self.rate is None:
            return dnstli2.rate.unlimited
        try:
            return self.rate.blkq_incr_limit([ blkq_env2RateParamsUser(env) ])
        except dnstli2.rate.RedisConnError:
            print("Got RedisConnError - using unlimited quota")
            return dnstli2.rate.unlimited
        except dnstli2.rate.LimitExceeded:
            log_rate_limit_exception(env)
            raise dnstli2.response.RateLimitExceeded
        except dnstli2.rate.BurstLimitExceeded:
            log_rate_limit_exception(env)
            raise dnstli2.response.BurstRateExceeded
        except dnstli2.rate.QuotaExpired:
            log_rate_quota_expired(env)
            raise dnstli2.response.QuotaExpired
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

empty_policy = object()
empty_acl = object()

# check if the REMOTE_ADDR is in our acl, if those are set.
# raise error if a problem and want_raise = True.
def check_acl(env, want_raise=True):
    if not env.get('dnstli.policy'):
        return empty_policy
    if not env['dnstli.policy'].get('addresses'):
        return empty_acl
    try:
        remote_ip = ipaddress.ip_address(env['REMOTE_ADDR'])
    except:
        # weird -- nginx/uwsgi passed us an invalid IP
        raise  dnstli2.response.InternalServerError
    for acl in itertools.chain(
            conf.get('admin_addresses', []),
            env['dnstli.policy']['addresses'],
            ):
        # check if REMOTE_ADDR is equal or inside the IP or CIDR
        # block specified by acl
        if remote_ip in ipaddress.ip_network(acl):
            return acl
    else:                       # nothing matched
        if want_raise:
            raise dnstli2.response.RemoteAddressUnauthorized
        return None

def check_method_args(env, method, args):
    """ will throw if sees a problem.  it will return when it thinks the query may be ok. """

    if env['dnstli.min_api_version'] > dnstli2.conf.MY_API_VERSION:
        raise dnstli2.response.APIVersionMismatch
    if env['dnstli.max_api_version'] < dnstli2.conf.MY_API_VERSION:
        raise dnstli2.response.APIVersionMismatch

    if (method == 'lookup' or method == 'summarize' or method == 'rate_limit') and not args:
        return                  # existing code in Lookup:__iter__ forms a 200 OK with no results for this

    qs = urllib.parse.parse_qs(env.get('QUERY_STRING',''))
    try:
        limit = int(qs['limit'][0])
        if limit < 0:
            raise dnstli2.response.BadRequest
    except ValueError:
        raise dnstli2.response.BadRequest
    except KeyError:
        pass                    # it is optional, so allow to not be there

    try:
        index, index_type, args = args.split('/', 2)
    except:
        raise dnstli2.response.BadRequest

    if index == 'rrset':
        if index_type not in ('name', 'raw'):
            raise dnstli2.response.BadRequest
        # could parse and check more: if rrtype is valid, for raw if unhexlify is ok
    elif index == 'rdata':
        if index_type not in ('name', 'ip', 'ip6', 'raw'):
            raise dnstli2.response.BadRequest
        # could parse and check more: if rrtype is valid, for raw if unhexlify is ok
    else:                       # unhandled index type
        raise dnstli2.response.BadRequest

    # survived without raising an error -- query is maybe ok

def setEnvFromBurstSizeAndWindow(env, burst_size, burst_window):
    """ Sets the burst size and window in env, only if either both are None or both are not None.
        Otherwise sets both to None.
    """
    if (burst_size is None) == (burst_window is None):
        env['dnstli.burst_size'] = burst_size
        env['dnstli.burst_window'] = burst_window
    else:
        print("Warning burst_size and burst_window are not either both None or both not None")
        env['dnstli.burst_size'] = None
        env['dnstli.burst_window'] = None


def getResultsLimit(env):
    """ Return the effective results limit """

    # Note res_limit_max must be set in the dnstli config or the program won't start
    program_max = conf['res_limit_max']
    try:
        # if there is an res_limit_override value in the per-subscription config,
        # then if it is an integer, it gives the maximum number of results to return.
        return min(int(env.get('dnstli.res_limit_override', program_max)), program_max)
    except:
        return program_max

def getOffsetMax(env):
    """ Return the effective offset max limit """

    # Note offset_max must be set in the dnstli config or the program won't start
    offset_max = conf['offset_max']
    if env.get('dnstli.res_limit_override'):
        return 'n/a'             # indicate offset not allowed
    else:
        return offset_max

class noauth(ratelimitmixin):
    def __init__(self, app):
        self.app = app

    def __call__(self, env, start_response):
        try:
            method, args = dnstli2.util.get_method_args(env)
            if method == 'ping':
                headers = [('Content-Type', 'application/x-ndjson')]
                start_response(dnstli2.response.Raw200Status, headers)
                return dnstli2.response.pingOk

            env['dnstli.authtype'] = 'open'
            env['dnstli.authuser'] = ''
            env['dnstli.authgroup'] = None
            env['dnstli.policy'] = None
            env['dnstli.limit'] = None
            env['dnstli.period'] = None
            env['dnstli.group_limit'] = None
            env['dnstli.group_period'] = None
            env['dnstli.res_limit_override'] = None
            setEnvFromBurstSizeAndWindow(env, burst_size = None, burst_window = None)

            if method == 'lookup' and args == 'auth':
                headers = [('Content-Type', 'text/plain')]
                start_response(dnstli2.response.AuthSuccess.status, headers)
                return [dnstli2.response.AuthSuccess.message]

            return self.app(env, start_response)
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

class apikey(ratelimitmixin):
    def __init__(self, app):
        self.app = app
        self.rate = get_rate_limiter()
        # self.last_BadAPIKey is a one-element cache of the last bad key we got.  Due to repeated bad keys,
        # we log a lot.  Don't log if the previous key was the same as this one
        self.last_BadAPIKey = None

    def load_key(self, env):
        if not 'HTTP_X_API_KEY' in env:
            raise dnstli2.response.MissingAPIKey

        try:
            api_key = env['HTTP_X_API_KEY']
            if apikey_pattern.match(api_key):
                key_fname = os.path.join(dnstli2.conf.API_DIR, api_key)
                #$$$ maybe cache based upon the last-mod-time to avoid reopening and parsing it for every query?
                with open(key_fname) as f:
                    keyConfDict = json.load(f)

                env['dnstli.authtype'] = 'api'
                env['dnstli.authuser'] = keyConfDict['username']
                env['dnstli.authgroup'] = keyConfDict.get('group')
                env['dnstli.policy'] = keyConfDict.get('policy')
                env['dnstli.res_limit_override'] = keyConfDict.get('res_limit_override')
                setEnvFromBurstSizeAndWindow(env, keyConfDict.get('burst_size'), keyConfDict.get('burst_window'))

                # Default the min api version to 1 and max to 9999
                env['dnstli.min_api_version'] = keyConfDict.get('min_api_version', 1)
                env['dnstli.max_api_version'] = keyConfDict.get('max_api_version', 9999)

                # Determine if auto-replenishing or block quota based.
                blkq_split = keyConfDict.get('blkq_split')

                if blkq_split is None:
                    # auto-replenishing quota
                    env['dnstli.limit'] = keyConfDict.get('limit')
                    env['dnstli.period'] = keyConfDict.get('period')
                    env['dnstli.group_limit'] = keyConfDict.get('group_limit')
                    env['dnstli.group_period'] = keyConfDict.get('group_period')
                else:
                    # block quota
                    env['dnstli.blkq_split'] = blkq_split
                    env['dnstli.blkq_baseline_used'] = keyConfDict.get('blkq_baseline_used')
                    env['dnstli.blkq_expires'] = keyConfDict.get('blkq_expires')
            else:
                # Note regarding these print statements: They output to /var/log/syslog (or messages).  We are
                # not using log(NOTICE, ...), which goes to /var/log/dnsdb/dnsdb.notice because that log is
                # mined by the brittle dnsdblog2erp script when it pushes the logs into postgres. log(DEBUG
                # etc are not by default configured by rsyslogd to be actually logged. So a print statement is
                # the only "safe" and default way to get a log message on disk
                if self.last_BadAPIKey == None or self.last_BadAPIKey != api_key:
                    print("BadAPIKey: does not match api key pattern", api_key, "from IP", env['REMOTE_ADDR'])
                self.last_BadAPIKey = api_key
                raise dnstli2.response.BadAPIKey
        except IOError as e:
            if e.errno == errno.ENOENT:
                # implement a one element memory of the last bad API key logged
                if self.last_BadAPIKey == None or self.last_BadAPIKey != api_key:
                    print("BadAPIKey: no key registered for key", api_key, "from IP", env['REMOTE_ADDR'])
                self.last_BadAPIKey = api_key
                raise dnstli2.response.BadAPIKey
            else:
                print("BadAPIKey: other IOError for key", api_key, "from IP", env['REMOTE_ADDR'])
                log_exception()
                raise dnstli2.response.BadAPIKey
        except Exception as e:
            # since ...BadAPIKey is an instance, not a pure class, can't catch it directly
            if e is dnstli2.response.BadAPIKey:
                raise
            print("BadAPIKey: other Exception for key", api_key, "from IP", env['REMOTE_ADDR'])
            log_exception()
            raise dnstli2.response.BadAPIKey

    def __call__(self, env, start_response):
        try:
            def start_response_add_limit_headers(status, headers, exc_info=None):
                ''' the enclosing __call__ function creates robj for me with rate info '''
                headers_cpy = list(headers)
                headers_cpy.append(('X-RateLimit-Limit', str(robj.limit)))
                headers_cpy.append(('X-RateLimit-Remaining', str(robj.remaining)))
                headers_cpy.append(('X-RateLimit-Reset', str(robj.reset)))
                if robj.expires != None:
                    headers_cpy.append(('X-RateLimit-Expires', str(robj.expires)))

                return start_response(status, headers_cpy, exc_info)
                # end of start_response_add_limit_headers

            try:
                method, args = dnstli2.util.get_method_args(env)

                # if all or part of method is a ping then ignore rest of the method
                if method[:5] == 'ping?' or method[:5] == 'ping/' or method == 'ping':
                    retCT = dnstli2.response.content_type_to_return(env,
                                                                    dnstli2.response.allowed_query_content_types,
                                                                    dnstli2.response.preferred_query_content_type)

                    headers = [('Content-Type', retCT)]
                    start_response(dnstli2.response.Raw200Status, headers)
                    return dnstli2.response.pingOk

                self.load_key(env)
                check_acl(env, want_raise=True) # will raise if our IP is not authorized

                check_method_args(env, method, args) # will throw if sees a problem

                if not conf.get('export_installation'):
                    if method == 'rate_limit':
                        retCT = dnstli2.response.content_type_to_return(env,
                                                                        dnstli2.response.allowed_query_content_types,
                                                                        dnstli2.response.preferred_query_content_type)

                        # rate_limit request does not count against the quota

                        if env.get('dnstli.blkq_split') is None:
                            # auto-replenishing quota
                            robj = self.autoq_get_limit(env)
                        else:
                            # block quota
                            robj = self.blkq_get_limit(env)

                        headers = [('Content-Type', retCT)]
                        start_response_add_limit_headers(dnstli2.response.Raw200Status, headers) # uses robj

                        # convert to a dict to get json key names too
                        rate_dict = {
                            'limit': robj.limit,
                            'remaining': robj.remaining,
                            'reset': robj.reset,
                            'results_max': getResultsLimit(env),
                            'offset_max': getOffsetMax(env)
                        }
                        if robj.expires != None:
                            rate_dict['expires'] = robj.expires
                        # add burst_size and burst_window if defined
                        if env.get('dnstli.burst_size') and robj.limit != 'unlimited':
                            rate_dict['burst_size'] = env.get('dnstli.burst_size')
                            rate_dict['burst_window'] = env.get('dnstli.burst_window')

                        return (json.dumps({"rate" : rate_dict}, separators=(',', ':')) + '\n').encode('utf-8')

                    # If limit exceeded, incr_limit will raise a RateLimitExceeded or BurstRateExceeded error
                    # and we'll return an HTTP error
                    if env.get('dnstli.blkq_split') is None:
                        robj = self.autoq_incr_limit(env)
                    else:
                        robj = self.blkq_incr_limit(env)
                        # self.app is where the real query processing takes place.
                        # It calls api.py:DNSTLI:__call__() which creates a api.py:Lookup() object
                    return self.app(env, start_response_add_limit_headers)
                else:           # export_installation -- no rate limits
                    if method == 'rate_limit':
                        headers = [('Content-Type', 'text/plain')]
                        start_response(dnstli2.response.NoRateLimits.status, headers)
                        return [dnstli2.response.NoRateLimits.message]
                    # self.app is where the real query processing takes place.
                    # It calls api.py:DNSTLI:__call__() which creates a api.py:Lookup() object
                    return self.app(env, start_response)
            except dnstli2.response.HTTPResponse as res:
                # Include the rate limit headers in the response only for the
                # 401 unauthorized or 429 too many requests errors.
                if res.status is dnstli2.response.Raw401Status or res.status is dnstli2.response.Raw429Status:
                    if env.get('dnstli.blkq_split') is None:
                        # auto-replenishing quota
                        robj = self.autoq_get_limit(env)
                    else:
                        # block quota
                        robj = self.blkq_get_limit(env)
                    start_response_add_limit_headers(res.status, res.headers)
                else:
                    start_response(res.status, res.headers)
                return [res.message]
        except Exception:
            log_exception()
            raise dnstli2.response.InternalServerError

